;; $Id: html-stream.scm 865 2004-07-19 18:07:45Z azul $
;;
;; This file is in the public domain and may be reproduced or copied without
;; permission from its author.  Citation of the source is appreciated.
;;
;; Alejandro Forero Cuervo <bachue@bachue.com>
;;
;; Tiny, functional and highly optimized HTML generator, generating srfi-40
;; streams.
;;
;; Documentation is available in HTML format.
;;
;; Newer versions might be available at:
;;
;;    http://anonymous:@afc.no-ip.info:8000/svn/home/src/chicken-eggs/html-stream

; Example:
;
; (html-stream
;   (html
;     ((body bgcolor "#ffffff")
;        (p "See " ((a href url id "somelink") "page " (current-seconds) " (now)") "."))))
;
; Expands to:
;
; (stream-delay
;  (->stream-char
;   "<html><body bgcolor='#ffffff'><p>See <a href='"
;   (stream-delay
;    (->stream-char
;     url
;     (stream-delay
;      (->stream-char
;       "' id='somelink'>page "
;       (stream-delay
;        (->stream-char
;         (current-seconds)
;         (stream-delay
;          (->stream-char
;           " (now)</a>.</p></body></html>"
;           stream-null))))))))))

(declare (uses srfi-1 extras))
(require-extension srfi-40 stream-ext)

(define-macro (html-stream . body)

  (define (html-write info rest)
    `(stream-delay (->stream-char ,info ,rest)))

  (define (join-show el rest)
    (if (or (null? rest) (not (string? el)) (not (string? (car rest))))
      `(,el ,@rest)
      `(,(string-append el (car rest)) ,@(cdr rest))))

  (define (html-parse body rest)
    (cond
      ((null? body) rest)
      ((list? (car body))
       (html-list (car body) (html-parse (cdr body) rest) (car body)))
      (else (cons (car body) (html-parse (cdr body) rest)))))

  (define (html-list element rest original)
    (if (list? (car element))
      (or (and-let* ((func (tag-func (caar element))))
            (func (caar element) (cdar element) (cdr element) rest))
          (cons original rest))
      (html-list `((,(car element)) . ,(cdr element)) rest original)))

  (define (html-params ps rest)
    (cond
      ((null? ps) rest)
      ((null? (cdr ps)) (error "Missing value for param: " (car ps)))
      (else (cons* " " (symbol->string (car ps)) "='" (cadr ps) "'"
                 (html-params (cddr ps) rest)))))

  (define (html-normal close)
     (lambda (name params body rest)
       (cons* "<" (symbol->string name)
         (html-params params
           (cons (if close ">" "/>")
             (html-parse body
               (if close (cons* "</" (symbol->string name) ">" rest) rest)))))))

  (define (tag-func name)
    (and (symbol? name)
      (case name
        ((style html head title body p a b i ul ol li dl dt dd table tr td thead tbody h1 h2 h3 h4 h5 h6 div span small big script select option textarea center form pre) (html-normal #t))
        ((img br hr meta input) (html-normal #f))
        (else #f))))

  (fold-right html-write 'stream-null (fold-right join-show '() (html-parse body '()))))
